@@ -16,6 +16,8 @@ import matplotlib.patheffects as path_effects
 from matplotlib import cm
 
 import plotly.graph_objects
+from plotly.offline import plot as plot_offline
+from plotly.offline import plot_mpl
 from adjustText import adjust_text
 
 from .GenericFunctions import read, write
@@ -49,7 +51,7 @@ class VisualizationFunctions:
 
         return
 
-    def saveFigure(self, fig, saveDir, label = 'Figure', extension = 'png', dpi = 300, close = True):
+    def saveFigure(self, fig, saveDir, label = 'Figure', extension = 'png', dpi = 300, close = True, attemptSavingHTML = False):
 
         '''Function used internally to save and close figures
 
@@ -103,6 +105,12 @@ class VisualizationFunctions:
         else:
             print('Unsupported format. Figure not saved')
 
+        if attemptSavingHTML:
+            try:
+                plot_mpl(fig, filename=os.path.join(saveDir, label + '.html'), auto_open=False)
+            except Exception as exception:
+                print('Saving to iteractive HTML did not succeed')
+
         if close:
             try:
                 plt.close(fig)
@@ -112,7 +120,7 @@ class VisualizationFunctions:
 
         return
 
-    def makeHeatmapGeneExpressionPlot(self, df = None, genes = None, nameToAppend = 'heatmap', dpi = 300, extension = 'png'):
+    def makeHeatmapGeneExpressionPlot(self, df = None, genes = None, nameToAppend = 'heatmap', dpi = 300, extension = 'png', **kwargs):
 
         '''Make heatmap gene expression plot from a provided gene expression matrix.
 
@@ -203,11 +211,11 @@ class VisualizationFunctions:
 
         fig.tight_layout()
 
-        self.saveFigure(fig, self.saveDir, self.dataName + '_' + nameToAppend + '_genes', extension=extension, dpi=dpi)
+        self.saveFigure(fig, self.saveDir, self.dataName + '_' + nameToAppend + '_genes', extension=extension, dpi=dpi, **kwargs)
 
-        return
+        return fig
     
-    def makeMarkerExpressionPlot(self, fontscale = 0.75, dpi = 300, extension = 'png'):
+    def makeMarkerExpressionPlot(self, fontscale = 1., dpi = 300, extension = 'png', **kwargs):
 
         '''Produce image on marker genes and their expression on all clusters.
         Uses files generated by function DCS.Vote
@@ -320,7 +328,7 @@ class VisualizationFunctions:
             clusterNames = list(votingResults.values())
             clusterIndices = list(votingResults.keys())
 
-            ax.set_yticklabels(['Mean'] + [str(clusterNames[i]) + ' (' + str(clusterIndices[i]) + ')' for i in ORDER], rotation=0, fontsize=8*fontscale)
+            ax.set_yticklabels(['Mean'] + [str(clusterNames[i]) + ' (' + str(clusterIndices[i]) + ')' for i in ORDER], rotation=0, fontsize=6*fontscale)
             ax.set_xlim([-0.5,X_markers_cluster_means_transpose.shape[1] - 0.5])
             ax.set_ylim([-0.5,X_markers_cluster_means_transpose.shape[0] + 1 - 0.5])
 
@@ -372,11 +380,11 @@ class VisualizationFunctions:
             ax3.set_xlim([-0.5, df_markers_weighted.shape[1] - 0.5])
             ax3.set_ylim([-0.5, df_markers_weighted.shape[0] - 0.5])
 
-        self.saveFigure(fig, self.saveDir, self.dataName + '_marker_expression', extension=extension, dpi=dpi)
+        self.saveFigure(fig, self.saveDir, self.dataName + '_marker_expression', extension=extension, dpi=dpi, **kwargs)
 
-        return
+        return fig
 
-    def internalMakeMarkerSubplots(self, df, X_projection, hugo_cd_dict, NoFrameOnFigures = False, HideClusterLabels = False, outlineClusters = True, analyzeBy = 'cluster', saveSubDir = 'marker_subplots', dpi = 300, extension = 'png'):
+    def internalMakeMarkerSubplots(self, df, X_projection, hugo_cd_dict, NoFrameOnFigures = False, HideClusterLabels = False, outlineClusters = True, analyzeBy = 'cluster', saveSubDir = 'marker_subplots', dpi = 300, extension = 'png', **kwargs):
 
         '''Produce subplots on each marker and its expression on all clusters
 
@@ -470,7 +478,7 @@ class VisualizationFunctions:
                 ax.axis('off')
             fig.tight_layout()
             if self.saveDir is not None: 
-                self.saveFigure(fig, directory, '%s_%s_%s_%s' % (self.dataName,marker,suffix.replace(',','_').replace('/','_'),analyzeBy), extension=extension, dpi=dpi)
+                self.saveFigure(fig, directory, '%s_%s_%s_%s' % (self.dataName,marker,suffix.replace(',','_').replace('/','_'),analyzeBy), extension=extension, dpi=dpi, **kwargs)
 
                 print(marker, end=" ", flush=True)
 
@@ -502,7 +510,7 @@ class VisualizationFunctions:
 
         return
 
-    def makeAnnotationResultsMatrixPlot(self, dpi = 300, extension = 'png'):
+    def makeAnnotationResultsMatrixPlot(self, dpi = 300, extension = 'png', **kwargs):
 
         '''Produce voting results voting matrix plot
 
@@ -598,11 +606,11 @@ class VisualizationFunctions:
         axx.set_xlabel('Number of\ncells in clusters', fontsize=20)
         axx.set_ylim(-0.5, num_of_clusters - 0.5)
         
-        self.saveFigure(fig, self.saveDir, self.dataName + '_scores_matrix', extension=extension, dpi=dpi)
+        self.saveFigure(fig, self.saveDir, self.dataName + '_scores_matrix', extension=extension, dpi=dpi, **kwargs)
  
-        return 
+        return fig
 
-    def makeHistogramNullDistributionPlot(self, dpi = 600, extension = 'png'):
+    def makeHistogramNullDistributionPlot(self, dpi = 600, extension = 'png', **kwargs):
 
         '''Produce histogram plot of the voting null distributions
 
@@ -726,13 +734,13 @@ class VisualizationFunctions:
 
             print()
         
-        self.saveFigure(fig, self.saveDir, self.dataName + '_null_distributions', extension=extension, dpi=dpi)
+        self.saveFigure(fig, self.saveDir, self.dataName + '_null_distributions', extension=extension, dpi=dpi, **kwargs)
 
         matplotlib.rcParams['axes.linewidth'] = origWidth
 
-        return
+        return fig
 
-    def makeProjectionPlot(self, Xprojection, cellClusterIndexLabel, suffix = '', colormap = cm.jet, legend = True, labels = True, colorbar = False, fontsize = 10, plotNaNs = True, rightShift = 0.3, dpi = 300, extension = 'png'):
+    def makeProjectionPlot(self, Xprojection, cellClusterIndexLabel, suffix = '', colormap = cm.jet, legend = True, labels = True, colorbar = False, fontsize = 10, plotNaNs = True, rightShift = 0.3, dpi = 300, extension = 'png', **kwargs):
 
         '''Produce projection plot (2D layout) with a specified coloring scheme
 
@@ -854,11 +862,11 @@ class VisualizationFunctions:
         if colorbar:
             add_colorbar(fig, possible_cluster_labels, cmap=colormap, fontsize=fontsize)
 
-        self.saveFigure(fig, self.saveDir, '%s_clusters_%s' % (self.dataName, suffix), extension=extension, dpi=dpi)
+        self.saveFigure(fig, self.saveDir, '%s_clusters_%s' % (self.dataName, suffix), extension=extension, dpi=dpi, **kwargs)
 
-        return
+        return fig
 
-    def makeStackedBarplot(self, clusterName = None, legendStyle = False, includeLowQC = True, dpi = 300, extension = 'png'):
+    def makeStackedBarplot(self, clusterName = None, legendStyle = False, includeLowQC = True, dpi = 300, extension = 'png', **kwargs):
         
         '''Produce stacked barplot with cell fractions
 
@@ -992,13 +1000,13 @@ class VisualizationFunctions:
             fig.tight_layout()
 
         saveName = "%s_subclustering_stacked_barplot_%s" % (self.dataName, ('All cell clusters' if clusterName == None else clusterName).replace(' ', '_').replace('*', ''))
-        self.saveFigure(fig, self.saveDir, saveName, extension=extension, dpi=dpi)
+        self.saveFigure(fig, self.saveDir, saveName, extension=extension, dpi=dpi, **kwargs)
 
         print('Saved stacked bar plot: %s' % ('All cell clusters' if clusterName == None else clusterName))
 
-        return
+        return fig
     
-    def makeQualityControlHistogramPlot(self, subset, cutoff, plotPathAndName = None, N_bins = 100, mito = False, displayMeasures = True, precision = 4, quantilePlotCutoff = 0.99, dpi = 300, extension = 'png', fontScale = 1.5, includeTitle = False):
+    def makeQualityControlHistogramPlot(self, subset, cutoff, plotPathAndName = None, N_bins = 100, mito = False, displayMeasures = True, precision = 4, quantilePlotCutoff = 0.95, dpi = 300, extension = 'png', fontScale = 1.5, includeTitle = False, **kwargs):
 
         '''Function to calculate QC quality cutoff and visualize it on a histogram
 
@@ -1150,11 +1158,11 @@ class VisualizationFunctions:
                 
             adjust_text(texts)
 
-        self.saveFigure(fig, os.path.dirname(plotPathAndName), label=os.path.basename(plotPathAndName) + '_histogram', extension=extension, dpi=dpi)
+        self.saveFigure(fig, os.path.dirname(plotPathAndName), label=os.path.basename(plotPathAndName) + '_histogram', extension=extension, dpi=dpi, **kwargs)
 
-        return
+        return fig
 
-    def makeSankeyDiagram(self, df, colormapForIndex = None, colormapForColumns = None, linksColor = 'rgba(100,100,100,0.6)', title = '', interactive = False, quality = 4, nameAppend = '_Sankey_diagram'):
+    def makeSankeyDiagram(self, df, colormapForIndex = None, colormapForColumns = None, linksColor = 'rgba(100,100,100,0.6)', title = '', attemptSavingHTML = False, quality = 4, nameAppend = '_Sankey_diagram'):
 
         '''Make a Sankey diagram, also known as 'river plot' with two groups of nodes
 
@@ -1215,7 +1223,7 @@ class VisualizationFunctions:
             sources.extend([i] * len(df.loc[item]))
             targets.extend(list(range(len(df.index), len(df.index) + len(df.loc[item]))))
             values.extend([j for j in df.loc[item].values])
-            labels.extend([item[0] + ' -> ' + jtem[0] for jtem in df.loc[item].index])
+            #labels.extend([item + ' -> ' + jtem for jtem in df.loc[item].index])
 
         colorscales = [dict(label=label, colorscale=[[0, linksColor], [1, linksColor]]) for label in labels]
 
@@ -1225,25 +1233,26 @@ class VisualizationFunctions:
                     newColor = ','.join(nodeColors[sources[i]].split(',')[:3] + ['0.6)'])
                     colorscales[i] = dict(label=labels[i], colorscale=[[0, newColor], [1, newColor]])
 
+        fig = plotly.graph_objects.Figure(data=[plotly.graph_objects.Sankey(valueformat = '', valuesuffix = '',
+            node = dict(pad = 20, thickness = 40, line = dict(color = 'white', width = 0.5), label = nodeLabels, color = nodeColors,),
+            link = dict(source = sources, target = targets, value = values, label = labels, colorscales = colorscales, hoverinfo='all'))]) #line ={'color':'rgba(255,0,0,0.8)', 'width':0.1}
+
+        if not title is None:
+            fig.update_layout(title_text=title, font_size=10)
+
         try:
-            fig = plotly.graph_objects.Figure(data=[plotly.graph_objects.Sankey(valueformat = '', valuesuffix = '',
-                node = dict(pad = 20, thickness = 40, line = dict(color = 'white', width = 0.5), label = nodeLabels, color = nodeColors,),
-                link = dict(source = sources, target = targets, value = values, label = labels, colorscales = colorscales))]) #line ={'color':'rgba(255,0,0,0.8)', 'width':0.1}
-
-            if not title is None:
-                fig.update_layout(title_text=title, font_size=10)
-
             fig.write_image(os.path.join(self.saveDir, self.dataName + nameAppend + '.png'), width=500, height=500, scale=quality)
 
         except Exception as exception:
-            print(exception)
+            print('Cannot save static image (likely due to missing orca). Saving to interactive html')
+            attemptSavingHTML = True
 
-        if interactive:
-            fig.show()
+        if attemptSavingHTML:
+            plot_offline(fig, filename=os.path.join(self.saveDir, self.dataName + nameAppend + '.html'), auto_open=False)
 
-        return
+        return fig
 
-    def makePlotOfNewMarkers(self, df_marker_cell_type, df_new_marker_cell_type, dpi = 300, extension = 'png'):
+    def makePlotOfNewMarkers(self, df_marker_cell_type, df_new_marker_cell_type, dpi = 300, extension = 'png', **kwargs):
 
         '''Produce plot of the new markers extracted from the annotated clusters
 
@@ -1300,11 +1309,11 @@ class VisualizationFunctions:
                     ax.plot(xy.T[0], xy.T[1], 'ro', markeredgecolor='r', ms=1.0, markeredgewidth=0.2)
 
         #ax.set_title('Additional markers along with the overlapping part of the input (red)')
-        self.saveFigure(fig, self.saveDir, self.dataName + '_new_markers', extension=extension, dpi=dpi)
+        self.saveFigure(fig, self.saveDir, self.dataName + '_new_markers', extension=extension, dpi=dpi, **kwargs)
 
-        return None
+        return fig
 
-    def makeTtestPlot(self, df, dfp, label = None, reorder = True, p_value_cutoff = 0.05, dpi = 300, extension = 'png'):
+    def makeTtestPlot(self, df, dfp, label = None, reorder = True, p_value_cutoff = 0.05, dpi = 300, extension = 'png', **kwargs):
 
         '''Produce heatmap plot of t-test p-Values calculated gene-pair-wise
         from the annotated clusters.
@@ -1400,11 +1409,11 @@ class VisualizationFunctions:
 
         axisColor.set_yticklabels([np.round(dataMax,2), np.round(dataMin,2)])
 
-        self.saveFigure(fig, self.saveDir, self.dataName + '_ttest_%s' % (label.replace('\n', '_')), extension=extension, dpi=dpi)
+        self.saveFigure(fig, self.saveDir, self.dataName + '_ttest_%s' % (label.replace('\n', '_')), extension=extension, dpi=dpi, **kwargs)
 
-        return None
+        return fig
     
-    def makeCellMarkersPiePlot(self, type1, type2, df_marker_cell_type = 'all', nameToAppend = None, listUnexpressedMarkers = True, orthogonalSectorsShift = 0.1, rotationAngle = 0, dpi = 300, extension = 'png'):
+    def makeCellMarkersPiePlot(self, type1, type2, df_marker_cell_type = 'all', nameToAppend = None, listUnexpressedMarkers = True, orthogonalSectorsShift = 0.1, rotationAngle = 0, dpi = 300, extension = 'png', **kwargs):
 
         '''Make summary of markers comparison between two cell types.
 
@@ -1670,11 +1679,11 @@ class VisualizationFunctions:
                 color='k', fontsize=8, ha='center', va='center').set_path_effects([path_effects.Stroke(linewidth=1, foreground='blue'),path_effects.Normal()])
 
         saveName = 'Markers_of_%s_vs_%s_(%s)_%s' % (type1.replace('/',''), type2.replace('/',''), nameToAppend, additional_name)
-        self.saveFigure(fig, self.saveDir, self.dataName + saveName, extension=extension, dpi=dpi)
+        self.saveFigure(fig, self.saveDir, self.dataName + saveName, extension=extension, dpi=dpi, **kwargs)
 
-        return dict(zip(labels, list(sets)))
+        return dict(zip(labels, list(sets))), fig
 
-    def makeHopfieldPCplot(self, colormap = cm.hot_r, plotTrLines = False, clusterid = 1, trID = 0, axisOff = False, fontscale = 1., trPath = None, dpi = 300, extension = 'png'):
+    def makeHopfieldPCplot(self, colormap = cm.hot_r, plotTrLines = False, clusterid = 1, trID = 0, axisOff = False, fontscale = 1., trPath = None, dpi = 300, extension = 'png', **kwargs):
 
         '''Make radar plot of the attractors in their principal components coordinates
 
@@ -1816,11 +1825,11 @@ class VisualizationFunctions:
         for label in ylabels:
             ax.text(label._x, label._y, label._text, zorder=np.inf)
 
-        self.saveFigure(fig, self.saveDir, self.dataName + '_polar_%s'%(suffix), extension=extension, dpi=dpi)
+        self.saveFigure(fig, self.saveDir, self.dataName + '_polar_%s'%(suffix), extension=extension, dpi=dpi, **kwargs)
 
-        return
+        return fig
 
-    def makeHopfieldLandscapePlot(self, legend = False, labels = False, PCx = 0, PCy = 1, colorbar = True, fontsize = 10, plotMesh = True, plotAttractors = True, adjustText = True, axisOff = True, colorbarva = 0.75, colorbarha = 0.85, trPath = None, colormap = matplotlib.colors.LinearSegmentedColormap.from_list('cmap', [(1, 1, 1), (0, 1, 1), (0, 0, 1), (1, 0, 0)], N=1000), dpi = 300, extension = 'png'):
+    def makeHopfieldLandscapePlot(self, legend = False, labels = False, PCx = 0, PCy = 1, colorbar = True, fontsize = 10, plotMesh = True, plotAttractors = True, adjustText = True, axisOff = True, colorbarva = 0.75, colorbarha = 0.85, trPath = None, colormap = matplotlib.colors.LinearSegmentedColormap.from_list('cmap', [(1, 1, 1), (0, 1, 1), (0, 0, 1), (1, 0, 0)], N=1000), dpi = 300, extension = 'png', **kwargs):
 
         '''Make heatmap plot of the attractors in their two principal components coordinates
 
@@ -1989,6 +1998,6 @@ class VisualizationFunctions:
 
             add_colorbar(fig, [vmax, vmin], cmap=colormap, fontsize=fontsize)
 
-        self.saveFigure(fig, self.saveDir, self.dataName + '_energy_landscape_PC%s_vs_PC%s'%(PCy, PCx), extension=extension, dpi=dpi)
+        self.saveFigure(fig, self.saveDir, self.dataName + '_energy_landscape_PC%s_vs_PC%s'%(PCy, PCx), extension=extension, dpi=dpi, **kwargs)
 
-        return
+        return fig
